from devapp.app import app
import os, json, sys
from devapp.tools import FLG, filter_passwords, exists, project, write_file, read_file
from devapp.tools import define_flags
import inspect


class flags:
    autoshort = 'ad'
    short_maxlen = 5

    class pytest_autodoc:
        """
        Create docu pages of the current pytest function within the build dir, containing build flow charts

        We generate the ready made markdown pages, with links to the flow charts.
        Current format only mkdocs.

        We only write the specs (.src) for those charts yet - since this is run while pytesting.
        Producing charts is done later.
        """

        d = False


define_flags(flags)
from theming.formatting import markdown

T = markdown.Mkdocs


class Skip(Exception):
    pass


class S:
    """state - set for each test func's call to def build_pipeline"""

    n_mod = None
    n_func = None
    n_cls = None
    mod = None
    cls = None
    func = None
    fn_doc = None
    d_root = None


from devapp.tools import deindent


from inflection import humanize


details = lambda s, b: T.details % (s, b)


def incr_header_levels(s, min_levels):
    s = '\n' + s
    return s.replace('\n#', '\n' + min_levels * '#')


def add_doc(body):
    with open(S.fn_doc, 'a') as fd:
        fd.write(body)


source = lambda src: details('Source code', T.py % src)


def do_mod():
    def find_mod(fn):
        for k, v in sys.modules.items():
            if (getattr(v, '__file__', None) or '').endswith(fn):
                return v

    S.mod = mod = find_mod(S.n_mod)
    # fn = lambda n: DTB() + '/' + n
    if not getattr(mod, 'is_auto_doc_file'):
        raise Skip(S.n_mod)
    fn = S.mod.__file__
    S.d_root = S.d_root or project.root()
    f = fn.split(S.d_root, 1)[1][1:].rsplit('.py', 1)[0] + '.md'
    S.fn_doc = os.path.join(S.d_root + '/build/autodocs', f)
    app.info('Generating autodoc file', fn=S.fn_doc)
    b = mod.__doc__
    if not b.strip():
        b = '# %s' % mod.__name__
    b += source(read_file(fn))
    write_file(S.fn_doc, b, mkdir=1)


def do_cls():
    b = '\n\n## %s\n\n' % S.n_cls
    S.cls = getattr(S.mod, S.n_cls)
    b += deindent(S.cls.__doc__ or '')
    add_doc(b)


func_title = lambda fn: humanize(fn.split('_', 1)[1])


def do_func():
    l, H = (2, 'h2') if not S.n_cls else (3, 'h3')
    h = '#' * l
    title = func_title(S.n_func)
    # b = '<a href="#%s">&nbsp;.</a>' % S.n_func
    b = '\n\n' + h + ' ' + title + '\n'
    # b = '\n\n%s %s\n###### %s\n' % ('#' * l, title, S.n_func)
    p = S.cls if S.n_cls else S.mod
    S.func = getattr(p, S.n_func)
    b += incr_header_levels(deindent(S.func.__doc__ or ''), l + 1)
    # b += add_chart()
    # b += add_flow()
    b += source(inspect.getsource(S.func))
    add_doc(b)


def register_new_test_func(pyt):
    l = pyt.split('::')
    funcs = []
    mod = l[0]
    funcs.append(do_mod) if mod != S.n_mod else 0
    S.n_mod = mod
    cls = None
    if len(l) > 2:
        cls = l[1]
        funcs.append(do_cls) if cls != S.n_cls else 0
    S.n_cls = cls
    S.n_func = l[-1].replace('(call)', '').strip()
    funcs.append(do_func)
    return funcs


dirname = os.path.dirname


def gen_func_doc():

    """We generate the markdown docs while running pytest process
    - we generate graph specs and write flow files for out of process analysis

    """
    pyt = os.environ.get('PYTEST_CURRENT_TEST')
    if not pyt:
        app.die('This is not a pytest run - cannot generate build doc')
    return [f() for f in register_new_test_func(pyt)]



# GK: For dumping, json is well faster than simplejson.
# For loading, simplejson is faster.
# tried with big request dumps, and even we have
# bool(getattr(simplejson, '_speedups', False))
import sys
from json import dumps

from pygments import highlight
from pygments.lexers import JsonLexer, YamlLexer
from yaml import dump, safe_dump

PY3 = sys.version_info[0] > 2
basestring = str if PY3 else basestring

try:
    from pygments.formatters import (
        Terminal256Formatter,
        TerminalFormatter,
    )
except:
    from pygments.formatters.terminal import TerminalFormatter

    Terminal256Formatter = TerminalFormatter


# from pygments.styles import get_style_by_name

ysl = ytermf = jsl = ''
# formatters by style:
_fmts = {}

Style = {}


def _fmt(style):
    if style in {'dark', 'light'} or Terminal256Formatter == TerminalFormatter:
        return TerminalFormatter(bg={'light': 'light'}.get(style, 'dark'))
    return Terminal256Formatter(style=style)


def get_fmt(style):
    termf = _fmts.get(style)
    if not termf:
        termf = _fmts[style] = _fmt(style)
    return termf


# ready to make partials for changing the defaults:
def coljhighlight(
    s,
    style=None,
    indent=4,
    sort_keys=True,
    add_line_seps=True,
    # logger might be set to colors off, e.g. at no tty dest:
    colorize=True,
    # automatic indent only for long stuff?
    no_indent_len=0,
):

    global jsl
    if not jsl:
        jsl = JsonLexer()
    if not isinstance(s, basestring):
        if indent and no_indent_len:
            if len(str(s)) < no_indent_len:
                indent = None
        try:
            s = dumps(s, indent=indent, sort_keys=sort_keys, default=str)
        except Exception as ex:
            try:
                # the sort may fail: TypeError: '<' not supported between instances of 'int' and 'str'
                s = dumps(s, indent=indent, default=str)
            except Exception as ex:
                print('breakpoint set')
                breakpoint()
                keep_ctx = True

    if colorize:
        # we may be called by structlog, then with style (from
        # FLG.log_dev_coljson_style)
        # or direct
        style = Style.get('style', style)
        if not style:
            try:
                from devapp.app import FLG

                style = FLG.log_dev_coljson_style
            except:
                # use the 16 base colors and leave to the terminal palette how to render:
                style = 'dark'
            Style['style'] = style
        res = highlight(s, jsl, get_fmt(style))
    else:
        res = s

    if add_line_seps:
        res = res.replace('\\n', '\n')

    return res


def colyhighlight(s, style='colorful'):
    global ysl, ytermf
    if not ysl:
        ysl = YamlLexer()

    ytermf = get_fmt(style)
    io = cStringIO.StringIO()
    if not isinstance(s, basestring):
        try:
            s = safe_dump(s)
        except:
            s = dump(s, default_flow_style=False)
    highlight(s, ysl, ytermf, io)
    res = io.getvalue()
    io.close()
    return res



"""
Strategy to be robust:
    unshare -m
    then mount overlay

    any change: begin from start
    state is kept in environ's 'DA_LAYERS' variable

"""


import os
from json import dumps, loads

from devapp import FLG, app, flag, run_app

env = os.environ.get
is_dir = os.path.isdir
exists = os.path.exists

flag.string('bases', env('DA_DIR', ''), 'Base dirs')
flag.boolean('create', False, 'Create Layer')
flag.boolean('remove', False, 'Remove Layer')


flag.string(
    'layer',
    '',
    (
        'Directory or match in $DA_DIR/layers or absolute dir.',
        'This will be on top, catching read write ops.',
    ),
)

flag.string(
    'overlay',
    env('DA_DIR', ''),
    'Created overlay directory - this is the resulting dir',
)


def show():
    return loads(env('DA_LAYERS') or '{}')


def write_state(state):
    start_pid = have['start_pid']
    fn = app.workdir + 'state.%s' % start_pid
    if exists(fn):
        app.log.dbg('State already written', fn=fn)
    else:
        with open(app.workdir + 'state.%s' % start_pid, 'w') as fd:
            fd.write(dumps(have))
        app.log.dbg('State written', **state)


def create():
    """Create Layers"""
    have = show()
    if have:
        write_state(have)
    breakpoint()
    fail = os.system('unshare -m ')


def main():
    """Called from max d package"""
    if FLG.create and FLG.remove:
        app.die('Cannot do both')
    if not FLG.create and not FLG.remove:
        return show()
    return (create if FLG.create else remove)()


layer = lambda: run_app(main)



from functools import partial
from json import dumps

import structlog
from theming.colorhilite import coljhighlight as coljson
from theming.term import RESET, Cell, Theme, structlog_style

# from theming.colorhilite import colyhighlight as colyaml


class formatters:
    coljson = partial(coljson, indent=2, no_indent_len=100, style=Theme.pygments_style)
    coljson_no_colors = partial(
        coljson, indent=2, no_indent_len=0, style=Theme.pygments_style, colorize=False,
    )

    # its done anyway key val based to single line, so we indent:
    json = partial(dumps, indent=4, default=str)


def build_val_fmtter(key, cfg, CFG):
    """cfg: Parameters of the val_formatters. CFG: config of the class"""
    func = cfg.pop('formatter')
    func_no_col = func if CFG['colors'] else func + '_no_colors'
    func = getattr(formatters, func_no_col, getattr(formatters, func, None))
    if not func:
        raise Exception('Formatter not found: %s' % func)
    return partial(func, **cfg)


call = structlog.dev.ConsoleRenderer.__call__


class ThemeableConsoleRenderer(structlog.dev.ConsoleRenderer):
    """
    Features:
    - Coloring based on theming.term (semantic styles)
              and not colorama (colors only)
    - on the fly insertion of unknown levels
    - configurable formatting of specific values:

    """

    def __init__(self, **kw):
        self.cfg, sl_kw = kw, {}
        # extract structlog original kw:
        for k in 'pad_event', 'repr_native_str':
            if k in kw:
                sl_kw[k] = kw.pop(k)

        colors = kw.get('colors')
        if colors:
            ax_style, ls = structlog_style()
            _ = 'level_styles'
            kw[_] = kw.get(_) or ls

        structlog.dev.ConsoleRenderer.__init__(self, colors=False, **sl_kw)

        self.colorful = colors
        if colors:
            # now  set OUR theme colors into a copy of sl's PlainStyle:
            self._styles = ax_style(self._styles)
            self._level_to_color = ls

        self.setup_value_formatters()

    def setup_value_formatters(self):
        """
        Example for value formatting:
        1. configure available formatters by name
            e.g. 'longj' might be a name for the coljson formatter with
            config: indent=2
        2. assign event dict keys to those names -> values will be formatted
        3. if you are ok with defaults you can point to the formatter funcname:
        renderer = ThemeableConsoleRenderer(
            val_formatters = {'longj': {'formatter': 'coljson', 'indent': 2}}
            # 'longj' is configured,  'json' is used as is:
            fmt_vals = {'result': 'longj', 'foo': 'json'}
            )
        """
        req_val_formatters = list(self.cfg.get('fmt_vals', {}).values())
        vf = dict(
            (k, build_val_fmtter(k, dict(v), self.cfg))
            for k, v in self.cfg.get('val_formatters', {}).items()
        )
        for k in req_val_formatters - vf.keys():
            vf[k] = build_val_fmtter(k, {'formatter': k}, self.cfg)
        self._val_formatters = vf

    def __call__(self, _, level, ev):
        if self.colorful:
            # every thread shows in unique color:
            try:
                tn = ev.pop('thread', None)
            except Exception as ex:
                print('breakpoint set')
                breakpoint()
                keep_ctx = True
        by = ev.pop('by', None)

        # configured value formatter functions?
        # let them do their work here and replace ev vals by formatted ones:
        # have to intermeditatly put into a placeholder, later replace back
        # reason: sl would screw the formatting:
        fvs = self.cfg.get('fmt_vals')
        if fvs:
            phs = []
            m = formatted_vals = {}
            for k in fvs:
                v = ev.pop(k, None)
                if v is not None:
                    phs.append(RESET + self._val_formatters[fvs[k]](v))
                    ev[k] = '_sl_%s__' % len(phs)

        try:
            s1 = call(self, _, level, ev)
        except Exception as ex:
            # Auto-Adding of new log levels:
            # loglevel 'blather' was already seen...
            if isinstance(ex, KeyError):
                ltc = self._level_to_color
                if level not in ltc:
                    ltc[level] = ltc['error']
                    return
            # Should never happen. To find this in the code should you get this:
            print('!' * 100, 'ax_rx_log_error', ev, str(ex))
            return

        # replace back:
        if fvs:
            while phs:
                s1 = s1.replace('_sl_%s__' % len(phs), phs.pop())

        if not self.colorful or tn is None:
            return s1

        # Thread: a unique looking cell in the terminal matrix:
        symb = Cell.unique(tn)
        s2 = s1
        # by is a special key
        if by:
            by = (by + ' ').ljust(5)
            # loglevel in?
            l = s1.split('] ', 1)
            if len(l) == 2:
                s1 = l[0] + '] ' + by + l[1]
            else:
                s1 = by + s1
        s = '\x1b[2m' + symb + ' ' + s2
        return s



"""
Some functions for creating simple value charts.

Currently only one, not yet very generalized but might be already now
of some use, expecially when it's about measuring timings.
Not very performant, nested loops.

Might add other charts later.


The current one mode makes this:

< snip >

    Total Run Times

 pool_new
  parallel_poolsize_1: ||||||||||||||||||| (0.43)
  serial             : ||||||||||||||||||| (0.44)
  parallel_big_pool  : |||||||||| (0.24)
  serial_no_pool     : ||||||||||||||||||||||||||||||||||||| (0.84)
     (1 tick is 0.023)

 pool_established
  parallel_poolsize_1: |||||||||||||||||| (0.42)
  serial             : ||||||||||||||||||| (0.44)
  parallel_big_pool  : | (0.031)
  serial_no_pool     : ||||||||||||||||||||||||||||||||||||| (0.85)
     (1 tick is 0.023)
======================================================================
    Average Run Times

 pool_new
  parallel_poolsize_1: ||||||||||||||||||||||||||||||||||| (0.01)
  serial             : |||||||||||||||||||||||||||||||||||| (0.011)
  parallel_big_pool  : ||||||||||||||||||||||||||||||||||||| (0.011)
  serial_no_pool     : |||||||||||||||||||||||||||||||||| (0.01)
     (1 tick is 0.00029)

 pool_established
  parallel_poolsize_1: |||||||||||||||||||||||||||||||||| (0.01)
  serial             : |||||||||||||||||||||||||||||||||||| (0.011)
  parallel_big_pool  : ||||||||||||||||||||||||||||||||||||| (0.011)
  serial_no_pool     : |||||||||||||||||||||||||||||||||| (0.01)
     (1 tick is 0.0003)
======================================================================
< / snip >

out of this:

< snip >

{'pool_established': {'parallel_big_pool': (0.031485795974731445,
                                            0.011009448766708374),
                      'parallel_poolsize_1': (0.42257595062255859,
                                              0.010400831699371338),
                      'serial': (0.43528890609741211, 0.010847651958465576),
                      'serial_no_pool': (0.84693813323974609,
                                         0.010381799936294556)},
 'pool_new': {'parallel_big_pool': (0.24394702911376953,
                                    0.010902762413024902),
              'parallel_poolsize_1': (0.43367099761962891,
                                      0.010407018661499023),
              'serial': (0.44476079940795898, 0.010810500383377076),
              'serial_no_pool': (0.84292483329772949, 0.010311901569366455)}}
< /snip >

via this command:

from theming.charting import bar_chart
print bar_chart(my_map, mode ='0:Vals,1:Main',  val_descr_list = ['\tTotal Run Times', '\tAverage Run Times'])


"""


def get_val_str(val):
    """ return the value as double precision string with max len 7 """
    if val > 10 and val < 1000000:
        return int(val)
    return '%.2g' % val


def normalize_vals(v, max_width, show_vals=1):
    """
    Normalize the values to a given width.
    """
    # normalize to this length:
    if show_vals:
        show_vals = 7
    else:
        show_vals = 0
    max_val_normalized = max_width - show_vals
    mv = max(v)
    tick_val = mv / float(max_val_normalized)
    ret = []
    for val in v:
        s = get_val_str(val)
        ret.append((val, int(val / tick_val), s))
    return ret, tick_val, get_val_str(tick_val)


def bar_chart(
    m,
    out_fmt='ascii',
    mode=None,
    show_vals=1,
    val_descr_list=None,
    symbol='|',
    legendwidth='auto',
    chart_width=70,
):
    """
        Creates ascii charts of python objects like maps or lists
        on the console
        """
    if mode == '0:Vals,1:Main':
        # if m like: {'Main1': {'V1': [1,3,2]}, 'V2':[5, 1, 1], ...}
        # we will draw sth like:
        """
            Value 1:
                Main1
                -----
                V1: | (1)
                V2: ||||| (5)
                (...)
            Value 2:
                Main1
                -----
                V1: ||| (3)
            (...)
            """
        assert len(symbol) == 1
        if legendwidth == 'auto':
            legendwidth = 0
            for l in m.values()[0].keys():
                kl = len(l)
                if kl > legendwidth:
                    legendwidth = kl

        if not val_descr_list:
            val_descr_list = []
            # check out the length of values (3 in the example):
            vals = len(m.values()[0].values()[0])
            for i in range(0, vals):
                val_descr_list.append('Value %s' % i)

        # the chart is built into this variable:
        res = ''
        v_index = -1

        # now loop through the first display hirarchy:
        for vd in val_descr_list:
            # in this loop we show that value of the value lists:
            v_index += 1

            res += '=' * chart_width + '\n'
            # print value description:
            res += vd + '\n'

            for k, v in m.items():
                val_list = []
                for vt in v.values():
                    try:
                        val_list.append(vt[v_index])
                    except:
                        raise ValueError(
                            'value count is not matching for row %s \
                                    of series %s, index is %s'
                            % (str(vt), v, v_index)
                        )
                width = chart_width - legendwidth - 7
                if not show_vals:
                    # don't show ' ()':
                    width = width + 3
                normalized, tickval, tick_fmt = normalize_vals(
                    val_list, width, show_vals=show_vals
                )

                res += '\n ' + k + '\n'
                # k like 'Main1', v like {'V1': [1,3,4], 'V2': ...}
                ind = -1
                for k1, v1 in v.items():
                    legend = (k1 + ' ' * 100)[:legendwidth] + ': '
                    ind += 1
                    val, ticks, v_disp = normalized[ind]
                    v_disp = ' (%s)' % v_disp
                    # k1 like V1
                    if not show_vals:
                        v_disp = ''
                    res += '  ' + legend + symbol * ticks + '%s\n' % v_disp
                res += '     (1 tick is %s)\n' % tick_fmt
        return res + '=' * chart_width

    raise NotImplemented


if __name__ == '__main__':

    print(normalize_vals([0.1, 34, 2], 70))
    print(normalize_vals([1, 39879709784, 0.000000078782], 70, '%.5e'))
    print('\n\n\n')
    print('Chart test')

    from pprint import pprint

    m = {
        'FirstSeries': {'test1': [1.12312, 0.001, 123123], 'test2': [23, 3434, 4545]},
        'SecondSeries': {
            'test1': [1, 2, 12342430.240],
            'test2': [23234240.001, 123123, 2],
        },
    }
    print('map to chart is:')
    pprint(m)

    print('\n\nDefault arguments\n')
    print(bar_chart(m, mode='0:Vals,1:Main'))
    print('\n' * 5)
    print('\n\nMore complex arguments\n')
    print(
        bar_chart(
            m,
            symbol='*',
            mode='0:Vals,1:Main',
            val_descr_list=['Condition 1', 'Condition 2', 'Condition 3'],
            show_vals=1,
            chart_width=50,
        )
    )
    print('\n\nNo values, width 80')
    print(
        bar_chart(
            m,
            symbol='*',
            mode='0:Vals,1:Main',
            val_descr_list=['Condition 1', 'Condition 2', 'Condition 3'],
            show_vals=None,
            chart_width=80,
        )
    )



"""
Parser for a Type / Class Tree

After a call to build_type_hirarchy, the types are put into a hirarchy, then
the class parsing starts.

The main features this thing provides is to detect type derived superclasses and
rearanges of them as new classes within the defining one.

        class Foo(Bar, Baz) ->
        class Foo(Bar):
                class Baz(Baz)

where in this not yet built when doing this,
we are in its metacls __new__ yet.

later Foo.Baz can be getting special settings without influencing the other
Baz classes.

This allows for very small definition specs.
"""

# TODO: Check this: https://www.stat.washington.edu/~hoytak/code/treedict/

# -------------------------------------------------------------- Build the Tree
# py 2 version. py2 syntax is compat with py3, so py3 can override
# cls_with_meta
import os
import sys

# from tree_builder.py2 import cls_with_meta
from collections import OrderedDict


def sys_path_insert(fn):
    """ helper, often required in SPECs, which does import us only"""
    sys.path.insert(0, os.path.abspath(os.path.dirname(fn)))


try:
    is_str = lambda s, t=str: isinstance(s, t)
except:
    is_str = lambda s: isinstance(s, (str, bytes))  # v3


try:
    # have?
    from theming.pretty_print import dict_to_txt

    class ODict(OrderedDict):
        def __repr__(self):
            return dict_to_txt(self)


except:
    import json

    ODict = OrderedDict
    dict_to_txt = lambda a, **kw: json.dumps(a, indent=4, default=str)


def pretty(o):
    from tree_builder.render import to_dict

    return dict_to_txt(to_dict(o), fmt={'ax': 1})


class ObjKeyDict(dict):
    set = dict.__setitem__


have_tty = lambda: sys.stdin.isatty() and sys.stdout.isatty()


class UserInput:
    """Avoids repeatedly asking same questions over repeated spec runs
    Stores answered questions in DA_DIR/.stored_input
    """

    fn = lambda: os.environ.get('DA_DIR', 'no_da_dir') + '/.stored_input'
    # when loaded (oncee from fn) then this keeps all previous user input:
    # inp e.g. = {'env': {'registration_token': 'foo'}}
    inp = None

    @classmethod
    def get_stored_if_build(ui, val_pth):
        """We do not want to prevent parsing of spec if env vars are missing.

        Only when its a spec build we return None, resulting in questions on the tty
        (or failures if non interactive)
        """

        if ui.inp is None:
            ui.load()
        m = ui.inp
        # val_pth e.g. ['env', 'reg_token']:
        while val_pth:
            m = m.get(val_pth.pop(0)) or {}
        if m != {}:
            return m

        # are we a spec build? If not return a placeholder:
        from devapp.app import app_func
        from devapp.spec.build import build

        if app_func(inner=True) != build:
            # placholder, this is not a spec build:
            return 'Input Value(%s)' % '.'.join(val_pth)

    @classmethod
    def load(ui):
        if not os.path.exists(ui.fn()):
            ui.inp = {}
            return
        with open(ui.fn()) as fd:
            ui.inp = json.loads(fd.read())

    @classmethod
    def store(ui):
        """called by successfull spec.build, storing last inputs"""
        if ui.inp:
            with open(ui.fn(), 'w') as fd:
                fd.write(json.dumps(ui.inp, indent=2))


def Env(key, dflt=None, suggest=None):
    """env access bridge for specs - here we can in a build process
    parametrize the constructor query session
    """
    import os

    res = os.environ.get(key, dflt)
    if res is None:
        res = UserInput.get_stored_if_build(['env', key])
        if res is None:
            if have_tty():
                print('Require value for: %s' % key)
                print(
                    'Note: Specifying a command (format: json) will delay value acquisition to process start time and prevent putting the value into process environ.'
                )
                print('Example: ["cat", "/tmp/mypw"]')
                if suggest:
                    print('Suggested (enter to confirm): %s' % suggest)
                res = UserInput.inp.setdefault('env', {})[key] = (
                    input('Value for %s: ' % key) or suggest
                )

            if not res:
                raise Exception('Required but not provided: $%s' % key)

    return res


def mro(c):
    return c.mro()


class Reg:
    """
    stores global registries,to allow isolated load of more specs in one
    process (clients could pass maps as 'env', then not using os.environ)
    """

    def __init__(self, env):
        self.env = env
        self.mixins = {}
        self.hirarchy = []
        self.cls_counter = 0
        self.hir_postconfigured = 2
        self.hirarchy_built = False
        self.classes = ObjKeyDict({})


R = Reg(os.environ)


def reinit(env=os.environ):
    global R
    R = Reg(env)
    return R


def load_isolated(spec, env=None):
    """execing a spec file (as string), should not collide
    only problem I see is that T's _allowed childs could differ,
    have to make an exception for it

    Note: R
    """
    global R
    if env is None:
        env = os.environ
    R = Reg(env=env)
    R.hirarchy.append(T)
    exec(spec, globals())
    return R.Root


pystring = str


def out(*a):
    pass
    # print ' '.join([str(k) for k in a])


class MixinTypeMetaClass(type):
    def __new__(cls, name, bases, attrs):
        ncls = type.__new__(cls, name, bases, attrs)
        # ncls = super(MixinTypeMetaClass, cls).__new__(cls, name, bases, attrs)
        R.mixins.setdefault(ncls, [])
        return ncls


def out_clses(cs):
    r = ''
    for c in cs:
        r += ', name: ' + c.__name__
    return 'Classes: ' + r[1:]


class MT(object, metaclass=MixinTypeMetaClass):
    pass


def build_type_hirarchy(root):
    """ e.g. Project"""
    while 1:
        if R.hirarchy[0] != root:
            R.hirarchy.pop(0)
        else:
            break
    # that hack now is necessary since transcypt requires the base with the
    # metaclass as FIRST base. Our Py2/Py3 compliance hack (cls_with_meta)
    # inserts an intermediate class - with a constructor which would overwrite
    # the actual wanted one (and we don't have super in Transcrypt, so that we
    # could fix it in the intermediate's __init__).
    # In short: We have to bypass __init__ of the intermediate helper base class:
    pfh = globals().get('post_fix_hirarchy')
    pfh() if pfh else None
    out('have hirarchy', [n.__name__ for n in R.hirarchy])
    R.hirarchy_built = True  # close hirachy, following are specific classes


class HirarchyErr(Exception):
    pass


def check_allow_add(cls, sub):
    if cls.__name__ == 'xProject':
        import pdb

        pdb.set_trace()
    try:
        if not sub._hirarchy == cls._hirarchy + 1:
            for c in cls._allowed_childs:
                if sub.type == c.type:
                    return
            raise HirarchyErr("Can't add %s directly within %s" % (sub, cls))
    except Exception as ex:
        print(ex)
        import pdb

        pdb.set_trace()


# ----------------------------------------------------------------------------
def cast_str(v):
    for t in float, int:
        try:
            return t(v)
        except:
            pass
    if v.lower() == 'true':
        return True
    if v.lower() == 'false':
        return False
    return v


nil = '\x01'


class ItemGetterMetaType(type):
    """with env to attr mapping (for inheritance. => DONT do a dict here)"""

    def __getitem__(cls, k, dflt=None):
        if isinstance(k, str) and k.startswith('$'):
            'could be env var pointing to our var OR env var'
            # we have it directly?:
            k = k[1:]
            ck = getattr(cls, '_env_%s' % k, nil)
            if ck != nil:
                return ck
            k = R.env.get(k, nil)
            if k == nil:
                return dflt
        return getattr(cls, k, dflt)

    def __setitem__(cls, k, v):
        # support cls['foo,bar'] = 'bar', 'foo'
        if isinstance(k, str):
            if k.startswith('$'):
                setattr(cls, '_env_%s' % k[1:], v)
            else:
                setattr(cls, k, v)
        else:
            # n.Consul['is_server', 'port_http', 'port_dns'] = True, 8880, 53
            for k, v in zip(k, v):
                setattr(cls, k, v)

    def get(cls, k, dflt=None):
        return cls.__getitem__(k, dflt)

    def get_env(cls):
        r = ODict()
        for k in [i for i in dir(cls) if i.startswith('_env_')]:
            r[k[5:]] = getattr(cls, k)
        return r


class Group(object, metaclass=ItemGetterMetaType):
    pass


class TypeMetaClass(ItemGetterMetaType):
    def __new__(cls, name, bases, attrs):
        try:
            attrs = dict(attrs)  # 0
        except:
            pass
        attrs['name'] = name
        if not bases:
            bases = (object,)
        b0, orig_bases = bases[0], None
        if b0 in R.mixins and not R.hirarchy_built:
            raise Exception("First Base Class can't be a mixin", b0)

        if R.hirarchy_built:
            out('adding a class', name, bases)
            if not '_parent' in attrs and not hasattr(b0, '_parent'):
                while not b0 in R.hirarchy:
                    # a helper class (like class Node(Node, Role.Foo) and inner
                    # Node is
                    # R.hirarchy type -> act as if we'd defined all:
                    inh = R.classes.get(b0)
                    b0, bases = inh[0], inh + bases[1:]
                attrs['type'] = b0.__name__
                attrs['_cls'] = name
                attrs['_hir'] = R.hirarchy.index(b0)
                out('adding {0}({1})'.format(name, b0.__name__))
                bases, orig_bases, mxins, attrs = set_bases_as_attrs(
                    name, b0, bases, attrs
                )
        else:
            # out ('Registering type', name)
            attrs['type'] = name
            attrs.setdefault('_allowed_childs', [])

        R.cls_counter += 1
        attrs['_id'] = R.cls_counter
        # ncls = super(TypeMetaClass, cls).__new__(cls, name, bases, attrs)
        ncls = type.__new__(cls, name, bases, attrs)
        if orig_bases:
            out(ncls.__name__, 'orig_bases', out_clses(orig_bases))
        if R.hirarchy_built:
            # if name == 'NB':
            #    import pdb; pdb.set_trace()
            if '_parent' in attrs:
                out('returning', ncls.__name__)
                return ncls
        if orig_bases:
            R.classes.set(ncls, orig_bases)
            for m in mxins:
                if hasattr(m, 'func_name'):
                    m(ncls)
                else:
                    out(
                        'Adding %s to %s %s'
                        % (name, R.mixins[m]['orig'].__name__, m.__name__)
                    )
                    R.mixins[m]['members'].append(ncls)
            out('built class', ncls.__name__, out_clses(R.classes.get(ncls)))
            return ncls
        if not R.hirarchy_built:  # and len(R.hirarchy):
            if R.hirarchy:
                R.hirarchy[-1:][0]._allowed_childs.append(ncls)
            ncls._hirarchy = len(R.hirarchy)
        R.hirarchy.append(ncls)
        return ncls

    def __repr__(cls):
        p = getattr(cls, '_parent', None)
        if p:
            # an object within an exploded tree:
            return fidstr(cls)

        if isinstance(p, pystring):
            p = 'prebound(%s)' % p
        elif isinstance(p, type(None)):
            p = 'type'
        else:
            if not getattr(p, '_parent', None):
                p = ''
            else:
                p = ' [%s]' % p
        return '%s.%s.%s%s' % (
            cls.type,
            getattr(cls, 'name', ''),
            getattr(cls, '_id', 0),
            p,
        )


class T(object, metaclass=TypeMetaClass):
    descr = None

    @classmethod
    def all(cls, type, **kw):
        def filter(o, kw):
            for k, v in kw:
                if getattr(o, k, None) == v:
                    return True

        # depends on root._all_<type>s to be set in build time:
        res = [
            o for o in getattr(cls._parents[0], '_all_%ss' % type) if cls in o._parents
        ]
        if not kw:
            return res
        kw = list(kw.items())
        return [o for o in res if list(filter(o, kw))]


class AXCTreeObj(T):
    pass


# T = cls_with_meta(TypeMetaClass, {'descr': None})


def has_parent(c):
    return getattr(c, '_parent', None)


def descr(c):
    'up the mro to find any descr or __doc__'
    d = []
    for b in mro(c):
        if b in R.hirarchy:  # this is a type not a class
            break
        dd = getattr(b, 'descr', None)
        if dd and not dd in d:
            d.append(dd)
        else:
            dd = b.__doc__
            if dd and not dd in d:
                d.append(dd)
    return '.'.join(d)


def has_typ(c):
    try:
        return hasattr(c, 'type')
    except:
        pass


def add_to(*dests, **kw):
    'convenience wrapper for add_cls - from spec'
    dests = _list(dests)
    for parent in dests:
        for k, childs in list(kw.items()):
            for child in _list(childs):
                if not child.type == k:
                    raise Exception("Can't add: %s is no %s" % (child, k))
                add_cls(parent, child)


def add_cls(parent, child):
    """ called from the spec """
    check_allow_add(parent, child)
    out('    adding %s to %s' % (child, parent))
    # R.cls_counter += 1
    new_sub = type.__new__(
        TypeMetaClass,
        child.__name__,
        (child,),
        {'_parent': parent, '_id': R.cls_counter},
    )
    # new_sub = type(child.__name__, (child,), {'_parent': parent})
    setattr(parent, child.__name__, new_sub)
    return new_sub


def set_bases_as_attrs(name, b0, orig_bases, attrs):
    """ type hirarchy is built at this point
    Our job now is to pop those bases (and add them as subclasses which
    registered already).
    If registered as mixin then we create new R.mixins or resolve - see below
    """

    def add_local(b0, name, sub, parent_attrs, lms=None):
        """
        """
        msg = 'Error trying to add %s(%s) to %s: ' % (sub, name, b0)
        try:
            check_allow_add(b0, sub)
        except HirarchyErr:
            if not lms:
                raise Exception(msg + ' no mro based sub')
            for i in range(len(lms), 0, -1):
                l = lms[i - 1]
                try:
                    t = add_cls(l, sub)
                    return t
                except HirarchyErr:
                    lms.pop(i - 1)
            raise Exception(msg + 'check_allow_add failed')
        bn = sub.__name__
        out('    adding {0} to {1}'.format(sub.__name__, name))
        # parent_attrs[bn] = t = type(bn, (sub,), {'_parent': name})
        # won't work with R.mixins, is basically creating a class also for python=0:
        parent_attrs[bn] = t = type.__new__(
            TypeMetaClass, bn, (sub,), {'_parent': name}
        )
        parent_attrs.setdefault('_from_mro', []).append(t)
        out('parent_attrs', len(parent_attrs), list(parent_attrs.keys()))
        return t

    mxins = []

    def new_mixin(ncls, b0, orig_mixin):
        if ncls in R.mixins:
            raise Exception('mixin %s already defined' % ncls)
        R.mixins[ncls] = {'members_cls': b0, 'members': [], 'orig': orig_mixin}

    mx = R.mixins.get(b0)
    if mx:
        # class NBI(AXESS) where AXESS is a group
        # -> we set group='AXESS' and b0 to Service
        mxins.append(b0)  # this will add this new cls to b0's members,
        # we don't have it yet
        attrs[mx['orig'].__name__] = b0.__name__
        b0 = mx['members_cls']
        attrs['type'] = b0.type

    new_bases = [b0]
    lms = []  # last added mro based sub
    out('orig_bases', out_clses(orig_bases))
    out('orig_bases1', out_clses(orig_bases[1:]))
    for b in orig_bases[1:]:
        out('base', b.__name__)
        if b in R.mixins:
            # a mixin type in the bases: class AXESS(Service, Group)
            # -> we'll create a NEW mixin, but with a members_type,
            # e.g. Service
            # if such a members_cls is alredy set, then we resolve
            # e.g. class Foo(Role, AXESS), AXESS is such a mx-> resolve
            mx = R.mixins[b]
            if 'members_cls' in mx:
                out('    resolving all members of %s' % b.__name__)
                for m in mx['members']:
                    # lms = [add_local(b0, name, sub=m, parent_attrs=attrs)]
                    lms = [add_local(b0, name, m, attrs)]
            else:
                # create new
                mxins.append(lambda ncls: new_mixin(ncls, b0, b))
        elif b in R.hirarchy:
            out(
                'you should instantiate your inner classes from R.hirarchy type classes'
            )
            import pdb

            pdb.set_trace()
        elif R.classes.get(b):
            # lms.append(add_local(b0, name, sub=b, parent_attrs=attrs, lms=lms))
            lms.append(add_local(b0, name, b, attrs, lms))
        else:
            # normal superclass:
            new_bases.append(b)
    return tuple(new_bases), orig_bases, mxins, attrs


def allow(*what, **on):
    # def allow(what):
    # 'allow((this, that), (onthis, onthat))'
    # what, on = what[0], what[1]
    on = on['on'] if isinstance(on['on'], (list, tuple)) else (on['on'],)
    for o in on:
        for w in what:
            o._allowed_childs.append(w)


def _list(s):
    if s and isinstance(s, tuple) and isinstance(s[0], (tuple, list)):
        s = s[0]
    return list(s) if isinstance(s, (list, tuple)) else [s]


# ------------------------------------------------------------ Tools
def lazy(d, key, func_if_missing, args):
    """ a lazy setdefault effectively. Often useful"""
    if key in d:
        return d[key]
    d[key] = func_if_missing(*args)
    return d[key]


# ------------------------------------------------------------ Explore the Tree
def props(c):
    # keys = [k for k in dir(c) if not k.startswith('_')]
    l = [(k, getattr(c, k)) for k in dir(c) if not k.startswith('_')]
    return [i for i in l if not hasattr(i[1], 'func_name') and not i[0] == 'all']


simple_types = (float, int, bool, pystring)


def simple_props(c):
    res = [(k, v) for k, v in props(c) if isinstance(v, simple_types)]
    return res


struct_types = (float, int, bool, pystring, dict, list, tuple)


def struct_props(c):
    'still serializable but with dicts and lists'
    res = [(k, v) for k, v in props(c) if isinstance(v, struct_types)]
    return res


def walk_tree(cur, pre=None, post=None, match=None, res=None, cfg=None, level=0):
    """
    Recursing the class tree, starting from a given root node (cur).
    the result is mutated
    pre, post, match = callables
    """
    if not match:
        match = has_typ
    if res is None:
        res = []
    level += 1
    if pre:
        pre(cur, res, level, cfg)
    if not cfg or cfg.get('stop_recursion') != cur.type:
        for k in ordered_childs(cur, match):
            walk_tree(k, pre, post, match, res, cfg, level=level)
    if post:
        post(cur, res, level, cfg)
    return res


def setup(root):
    """ configure all prebound classes """
    cfg = {'hashes': [], 'root': root}
    out('walking tree', root.__name__)
    root._level = root._hir = 0
    root._parents = [root]

    # register (API): _all_<type>s, e.g. _all_clusters
    def type_list_n(c):
        return '_all_' + c.type.lower() + 's'

    for h in R.hirarchy:
        setattr(root, type_list_n(h), [])

    def pre(cur, res, level, cfg):
        root = cur._root = cfg['root']
        getattr(root, type_list_n(cur)).append(cur)
        if cur != root:
            cur._parents = list(cur._parent._parents)
            # c._cluster -> Cluster.foo
            for c in cur._parents:
                setattr(cur, '_%s' % c.type.lower(), c)

            cur._parents.append(cur)

        return unordered_childs(cur, has_typ, level)

    walk_tree(root, pre=pre, post=None, match=has_typ, cfg=cfg)
    R.hirarchy_built = R.hir_postconfigured
    root.descr = descr(root)
    R.Root = root
    root._hirarchy_list = R.hirarchy
    return root


create_tree = setup


def scale(c, into, n):
    for i0 in range(n):
        i = i0 + 1
        name = c.name + str(i)
        t = type(name, (c,), {})
        t.__doc__ = c.__doc__
        into[name] = t


def unordered_childs(parent, match, level=None):

    # if parent.name == 'NB':
    #    import pdb; pdb.set_trace()
    c = getattr(parent, '_childs', None)
    if c:
        return c
    childs = [(k, v) for k, v in props(parent) if match(v)]
    if not childs:
        _childs = childs
    else:
        # we sort class childs by mro order, e.g. NB(N, A, B, C) ->
        # _childs = [A, B, C] even if C._id < A._id since defined earlier:
        try:
            pmro = parent._from_mro
            childs.sort(key=lambda x: pmro.index(x[1]))
        except Exception as ex:
            # some parents have no _from_mro
            childs.sort(key=lambda x: x[1]._id)

        _childs = []
        for k, v in childs:
            # v = TypeMetaClass.__new__(TypeMetaClass, v.__name__, (v,), {})
            R.cls_counter += 1
            # v = TypeMetaClass.__new__(TypeMetaClass, v.__name__, (v,), {'_id': R.cls_counter})
            # if not isinstance(v._parent, basestring):
            #    import pdb; pdb.set_trace()
            #    v = v.__bases__[0]
            v = type.__new__(
                TypeMetaClass,
                v.__name__,
                (v,),
                {'_id': R.cls_counter, '_level': level, '_parent': parent},
            )
            setattr(parent, k, v)
            v.descr = descr(v)
            _childs.append(v)
    parent._childs = _childs
    # parent._childs_ordered = False
    return _childs


# spec API Functions:
def ordered_childs(parent, match):
    # match is always the same,thats why we can store the result:
    return unordered_childs(parent, match)


def full_id(c, sep=None):
    res = [o.name for o in c._parents]
    return sep.join(res) if sep else res


def assign(key, objs, vals):
    [setattr(_[0], key, _[1]) for _ in zip(objs, vals)]


def setter(key):
    def set(key, v, *o):
        for obj in o:
            setattr(obj, key, v)

    return lambda v, *o: set(key, v, *o)


# full id as string:
fidstr = lambda cls: '.'.join([c.name for c in cls._parents])


def get_spec_root(spec):
    for k in dir(spec):
        root = getattr(getattr(spec, k), '_root', None)
        if root:
            return root
    die('Could not determine spec root', spec)


if __name__ == '__main__':

    class Project(T):
        ''

    class Cluster(T):
        ''

    build_type_hirarchy(root=Project)

    class cluster(Group):
        class A(Cluster):
            'A cluster'

        class B(Cluster):
            'B cluster'

    class MyP(Project, cluster['A'], cluster.B):
        'Pro'

    setup(MyP)
    print(('have ', dir(MyP)))



def port_from_preset(p):
    pr = getattr(presets['ports'], p, None)
    if not pr:
        raise Exception('Did not find port for %s in presets' % p)
    return pr


# ----------------------------------------------------------------------------
from urllib.parse import urlsplit, parse_qsl


def parse_via(url, props=None, default_port=None, add_caps_str=False):
    """
    'mysql+pymysql://root@127.0.0.1:3306/diameap' to
    {'caps': ['mysql', 'pymysql'],
     'hostname': '127.0.0.1',
     'netloc': 'root@127.0.0.1:3306',
     'path': '/diameap',
     'port': 3306,
     'scheme': 'mysql+pymysql',
     'username': 'root'}

    dia:
    url = 'aaa://127.0.0.1:13868;transport=tcp;protocol=diameter'
    we treat like it would be: 'aaa+tcp+diameter://127.0.0.1:13868?transport=tcp&protocol=diameter'

    'aaa://127.0.0.1:13868;transport=tcp;protocol=diameter' ->
    {'caps': ['aaa', 'tcp', 'diameter'],
     'hostname': '127.0.0.1',
     'netloc': '127.0.0.1:13868',
     'port': 13868,
     'protocol': 'diameter',
     'scheme': 'aaa',
     'transport': 'tcp'}
    """
    from .__init__ import simple_props

    props = {} if props is None else props
    if isinstance(url, dict):
        if 'port' in url:
            url['port'] = port_from_preset(url['port'])
        return url
    if not url:
        url = 'ip://'
    if not '://' in url:
        url += '://'

    u = urlsplit(url)
    hostport = u.netloc
    semicolnfos = None
    if ';' in hostport:
        hostport, semicolnfos = hostport.split(';', 1)
        u = urlsplit(url.replace(';' + semicolnfos, ''))

    nl = u.netloc.split(':')
    if len(nl) == 2 and not nl[1].isdigit():
        nl[1] = default_port or str(port_from_preset(nl[1]))
        url = url.replace(u.netloc, ':'.join(nl))
        u = urlsplit(url)

    m = props
    m.update(dict([(k, v) for k, v in simple_props(u) if v]))
    if m.get('hostname') == '*':
        m['hostname'] = '0.0.0.0'
    if not m.get('scheme'):
        msg = 'Could not derive Scheme: %s' % url
        if '_' in url.split('://', 1)[0]:
            msg += '. Scheme may not contain underscores. Use minus'
        raise Exception(msg)
    s = m['scheme']
    c = m['caps'] = s.split('+')
    if 'https' in c:
        m['ssl'] = True
        m['caps'][c.index('https')] = 'http'
    elif 'aaas' in c:  # diameter
        m['ssl'] = True
        m['caps'][c.index('aaas')] = 'aaa'

    m.update(parse_qsl(m.pop('query', {})))

    # dia:
    if semicolnfos:
        s = parse_qsl(semicolnfos.replace(';', '&'))
        m.update(s)
        m['caps'].extend([i[1] for i in s])
    if add_caps_str:
        m['caps_str'] = '+'.join(m['caps'])
    return m


isconn = lambda c: getattr(c, '__isconn__', False)


class Conn:
    __isconn__ = True
    via, src, next, sink = None, None, None, None

    @classmethod
    def show(c, ind=2):
        if not hasattr(c, 'src'):
            return
        via = c.via['caps'][0]
        sink = c.sink.name if c.sink else '-'
        # print('foo' * 1000)
        print(
            '%s src: %s, via: %s, sink: %s, next: %s'
            % (' ' * ind, c.src.name, via, sink, c.next.src.name if c.next else '-',)
        )
        ind = ind + 2
        c.next.show(ind) if c.next else None

    @classmethod
    def to(cls, sink, via=None):
        if via is None:
            via = cls.end().via
        if cls.next:
            c = cls.next.to(sink, via)
            return cls
        else:
            if not cls.sink:
                # we build a new one with same source, was given as
                # connect(foo).to(bar) -> first one is irrelevant
                connection_classes.pop()
            c = connect(cls.sink or cls.src, to=sink, via=via)
            if not cls.sink:
                return c
            cls.next = c
            cls.sink = None
        return cls

    @classmethod
    def end(cls):
        return cls if cls.sink else cls.next.end() if cls.next else cls


connection_classes = []


def connect(frm, to=None, via=None, **props):
    end = frm.end() if isconn(frm) else None
    if end and via is None:
        via = end.via
    via = parse_via(via, props)

    class Connection(Conn):
        pass

    C = Connection
    C.src = end.sink if end else frm
    C.sink = to
    C.next = None
    C.via = via
    if isconn(to):
        C.next = to
        C.sink = None
    if isconn(frm):
        end.next = C
        end.sink = None
        return frm
    connection_classes.append(C)
    return C


all_clses_by_conn = {}
from time import time


def map_connections_to_all_classes(root):
    """
    "connect(Nginx, Auth)" could mean tons of specific connections - since
    src and sink can be parent classes for many specific ones. Here we build
    out the all_clses_by_conn map, mapping all conn. objs to their affected
    sources and sinks
    """

    t1 = time()

    def add(conn, to, attr):
        if not hasattr(to, '_hirarchy'):
            # just e.g. a string
            return
        setattr(to, attr, getattr(to, attr, ()) + (conn,))

    for c in connection_classes:
        all_clses_by_conn[c] = {'src': [], 'sink': []}
        # if c.src._id == 22:
        #    import pdb; pdb.set_trace()
        add(conn=c, to=c.src, attr='_src')
        add(conn=c, to=c.sink or c.next.src, attr='_sink')

    # now we need all SPECIFIC classes for a connection -> we walk the tree:
    def set_conn(cls):
        for t in 'src', 'sink':
            conns = getattr(cls, '_' + t, ())
            for c in conns:
                all_clses_by_conn[c][t].append(cls)

    def walk(cur):
        set_conn(cur)
        for c in cur._childs:
            walk(c)

    walk(root)
    # print time() - t1


fullid = lambda cls: [c.name for c in cls._parents]


def conns_by_cls_dependent(cls):
    """We go through all conns to us (is_sink)
    and pop their resulting src conns
    A daemon like nginx should then have no more source conns, since outgoing
    are resulting from incomming.
    """
    outs = list(getattr(cls, '_src', ()))
    ins = list(getattr(cls, '_sink', ()))
    d_ins = []
    for c in ins:
        m = {'cls': all_clses_by_conn[c]['src'], 'via': c.via}
        n = c.next
        if n:
            outs.remove(n)
            m['sink'] = sink(n)
        d_ins.append(m)
    r = {'out': [sink(s) for s in outs], 'in': d_ins, 'cls': cls}
    return r


def sink(n):
    return {'via': n.via, 'cls': all_clses_by_conn[n]['sink']}


def find_connected_classes(cls, into):
    outs = list(getattr(cls, '_src', ()))
    ins = list(getattr(cls, '_sink', ()))
    if not outs and not ins:
        return
    for conns, t in (outs, 'sink'), (ins, 'src'):
        for c in conns:
            all = all_clses_by_conn[c][t]
            [into.append(a) for a in all if not a in into]


is_str = lambda s: isinstance(s, str)
is_ext_src = lambda s: s in ('*',)
is_end_sink = lambda s: is_str(s)
is_start_source = lambda s: len(s.split('.')) == 4 or is_ext_src(s)


def link(*what, **kw):
    'first arg may be name, else name is ""'
    name = ''
    if len(what) > 0:
        if is_str(what[0]) and not is_start_source(what[0]):
            name = what[0]
            what = what[1:]
    if len(what) == 0:
        raise Exception(kw, 'You did not specificy what to link (i.e. a client)')
    sinks = kw.pop('sink', 0)
    if sinks == 0:
        raise Exception('Link: "sink" is a required argument', what, kw)
    if not isinstance(sinks, (tuple, list)):
        sinks = (sinks,)
    root = what[0]._root if is_end_sink(sinks[0]) else sinks[0]._root

    links = root._links = getattr(root, '_links', [])

    # populate:
    [setattr(s, '_srcs', getattr(s, '_srcs', [])) for s in sinks if not is_end_sink(s)]
    for c in what:  # src -> sink
        # c could be e.g. '0.0.0.0', for a standalone sink
        standalone_sink = True
        cid = c
        if hasattr(c, '_parent'):
            c['_sinks'] = getattr(c, '_sinks', [])
            standalone_sink = False
            cid = c._id
        via = kw.pop('via')
        for s in sinks:
            link = {'src': c, 'sink': s, 'name': name}
            if via:
                link.update(parse_via(via))
            link['id'] = (cid, s._id if not is_end_sink(s) else s)
            link.update(kw)
            links.append(link)
            if not standalone_sink:
                c._sinks.append(link)
            if not is_end_sink(s):
                s._srcs.append(link)


def set_connections(spec, fid, attr='connections'):
    root = getattr(spec, fid.split('.', 1)[0])
    n = '_have_%s' % attr
    if getattr(root, n, None):
        return
    conns = [(c, getattr(root, c)) for c in dir(root)]
    conns = [c for c in conns if getattr(c[1], '__isconn__', False)]
    [set_links(root, conn[0], conn[1], attr) for conn in conns]
    setattr(root, n, 1)


no_conn = lambda: {'is_sink': [], 'is_src': []}


def get_conns(c, type, attr):
    conns = getattr(c, attr, None)
    if not conns:
        conns = no_conn()
        setattr(c, attr, conns)
    return conns[type]


def set_links(root, name, conn, attr):
    while True:
        conn.name = name
        src, sink, via = conn.src, conn.sink or conn.next.src, conn.via
        src_conns, sink_conns = (
            get_conns(src, 'is_src', attr),
            get_conns(sink, 'is_sink', attr),
        )
        cm = {
            'via': conn.via,
            'name': name,
            'src': src.name,
            'sink': sink.name,
            'obj': conn,
        }
        src_conns.append(cm)
        sink_conns.append(cm)
        if conn.sink:
            return
        conn = conn.next


presets = {}


def add_link(ports=None):
    # In Specs: L = add_link(ports=Ports) -> allows to use strings for ports
    presets['ports'] = ports or {}
    if not ports:
        return connect
    offs = getattr(ports, '_offset_', None) or 0
    if not str(offs).isdigit():
        raise Exception('No numeric port offset: %s' % off)
    offs = int(offs)
    s = lambda k, m: k.startswith(m)
    for k in dir(ports):
        if not s(k, 'ext_') and not s(k, '_'):
            p = int(getattr(ports, k))
            if p + offs > 655535 or p + offs < 10:
                raise Exception(
                    'Port must be 10 < port < 655535. have  %s' % (p + port)
                )
            setattr(ports, k, p + offs)
    return connect


def dump_all(root, fmt=None):
    map_connections_to_all_classes(root)
    r = {}
    for s in root._all_services:
        r[str(s)] = conns_by_cls_dependent(s)
    if fmt == 'json':
        import json

        r = json.dumps(r, indent=2, sort_keys=True, default=str)
    return r



"""
# Multi Purpose Flags

Additionally we have log_level, ... flags, defined in structlogging.sl
"""
from devapp.app import FLG, flag


def oauth():

    # e.g. gitlab, cloudfoundry
    flag.string('username', '', 'Username')
    flag.string('endpoint_name', 'default', 'You may have more endpoints')
    flag.string('endpoint', '', 'API Endpoint')

    def get_endpoint(example='https://example.gitlab.com'):
        import requests
        from devapp.spec.os_tools import env_get_interactive_write as envget

        def vld(k, url):
            try:
                if requests.get(url).status_code == 200:
                    return 'write'
            except:
                print('Not reachable')
                return

        return FLG.endpoint or envget('DA_URL_GITLAB', validate=vld, example=example)

    return get_endpoint
